"""Filter rules basing on host and device compiler names and versions.
"""
import io

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_job_coverage.util import (
    row_check_name,
    row_check_version,
    is_in_row,
    reason,
)

from packaging import version as pk_version
from typing import List, Tuple, Union, Optional
from typeguard import typechecked


def get_required_parameter() -> List[str]:
    """Returns a list of parameters which are required for using the filter defined by this module.
    Returns:
        List[str]: list of parameters
    """
    return [HOST_COMPILER, DEVICE_COMPILER]


@typechecked
def compiler_version_filter_typed(
    row: List[Union[Tuple[str, str], List[Tuple[str, str]]]],
    output: Optional[Union[io.StringIO, io.TextIOWrapper]] = None,
) -> bool:
    """Type checked version of compiler_version_filter(). Should be only used for
    testing or tooling. The type check adds a big overhead, which slows down
    pair-wise generator by the factor 30.

    Args:
        row (List[Union[Tuple[str, str], List[Tuple[str, str]]]]): Combination
        to verify. The row can contain up to all combination fields and at least
        two items.
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): Write
        additional information about filter decisions to the IO object
        (io.SringIO, sys.stdout, sys.stderr). If it is None no information is
        generated.

    Returns:
        bool: True, if combination is valid, otherwise False.
    """
    return compiler_version_filter(row, output)


def compiler_version_filter(
    row: List, output: Optional[Union[io.StringIO, io.TextIOWrapper]] = None
) -> bool:
    """Filter rules basing on host and device compiler names and versions.

    Args:
        row (List): Combination to verify. The row can contain
        up to all combination fields and at least two items.
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): Write
        additional information about filter decisions to the IO object
        (io.SringIO, sys.stdout, sys.stderr). If it is None no information is
        generated.

    Returns:
        bool: True, if combination is valid, otherwise False.
    """
    # if the same compiler is used for host and device code, it is not possible to use
    # different versions
    if (
        is_in_row(row, HOST_COMPILER)
        and is_in_row(row, DEVICE_COMPILER)
        and (
            # nvcc is a exception, because the host compiler is different to device compiler
            row_check_name(row, DEVICE_COMPILER, "!=", NVCC)
            and row[param_map[HOST_COMPILER]][VERSION]
            != row[param_map[DEVICE_COMPILER]][VERSION]
        )
    ):
        reason(
            output,
            "the host and device compilers must have the same version (except for nvcc)",
        )
        return False

    if row_check_name(row, DEVICE_COMPILER, "==", NVCC):
        cuda_sdk_version = 0
        cuda_host_compiler_version = 1

        # set the and lowest highest supported gcc version for nvcc
        if row_check_name(row, HOST_COMPILER, "==", GCC):
            combinations = [
                # it needs to be defined the CUDA SDK version, which supports
                # a new gcc version the first time
                # the latest CUDA SDK, also if it supports no new gcc version
                # (maximum_CUDA_SDK_version, "maximum_gcc_version")
                ("12.2", "12"),
                ("12.0", "12"),
                ("11.4", "11"),
                ("11.1", "10"),
                ("11.0", "9"),
                ("10.1", "8"),
                ("10.0", "7"),
            ]

            if pk_version.parse(
                row[param_map[DEVICE_COMPILER]][VERSION]
            ) <= pk_version.parse(combinations[0][cuda_sdk_version]):
                # check the maximum supported compiler version
                for combination in combinations:
                    if pk_version.parse(
                        row[param_map[DEVICE_COMPILER]][VERSION]
                    ) >= pk_version.parse(combination[cuda_sdk_version]):
                        if pk_version.parse(
                            row[param_map[HOST_COMPILER]][VERSION]
                        ) > pk_version.parse(combination[cuda_host_compiler_version]):
                            reason(
                                output,
                                f"nvcc-{row[param_map[DEVICE_COMPILER]][VERSION]} "
                                f"does not support gcc-{row[param_map[HOST_COMPILER]][VERSION]}",
                            )
                            return False
                        else:
                            break

                # since CUDA 11.4, the minimum supported GCC compiler is GCC 6
                if pk_version.parse(
                    row[param_map[DEVICE_COMPILER]][VERSION]
                ) >= pk_version.parse("11.4") and pk_version.parse(
                    row[param_map[HOST_COMPILER]][VERSION]
                ) < pk_version.parse(
                    "6"
                ):
                    reason(
                        output,
                        "as of CUDA 11.4 the minimum required GCC compiler is gcc-6",
                    )
                    return False

        if row_check_name(row, HOST_COMPILER, "==", CLANG):
            # disable clang as host compiler for nvcc 11.3 until 11.5
            if row_check_version(
                row, DEVICE_COMPILER, ">=", "11.3"
            ) and row_check_version(row, DEVICE_COMPILER, "<=", "11.5"):
                reason(
                    output,
                    "clang as host compiler is disabled for nvcc-11.3 to 11.5",
                )
                return False

            combinations = [
                # it needs to be defined the CUDA SDK version, which supports
                # a new clang version the first time
                # the latest CUDA SDK, also if it supports no new clang version
                # (maximum_CUDA_SDK_version, "maximum_clang_version")
                ("12.2", "15"),
                ("12.1", "15"),
                ("12.0", "14"),
                ("11.6", "13"),
                ("11.4", "12"),
                ("11.2", "11"),
                ("11.1", "10"),
                ("11.0", "9"),
                ("10.1", "8"),
                ("10.0", "6"),
            ]

            if pk_version.parse(
                row[param_map[DEVICE_COMPILER]][VERSION]
            ) <= pk_version.parse(combinations[0][cuda_sdk_version]):
                # set the and lowest highest supported clang version for nvcc
                # check the maximum supported compiler version
                for combination in combinations:
                    if pk_version.parse(
                        row[param_map[DEVICE_COMPILER]][VERSION]
                    ) >= pk_version.parse(combination[cuda_sdk_version]):
                        if pk_version.parse(
                            row[param_map[HOST_COMPILER]][VERSION]
                        ) > pk_version.parse(combination[cuda_host_compiler_version]):
                            reason(
                                output,
                                f"nvcc-{row[param_map[DEVICE_COMPILER]][VERSION]} "
                                f"does not support clang-{row[param_map[HOST_COMPILER]][VERSION]}",
                            )
                            return False
                        else:
                            break

    if row_check_name(row, DEVICE_COMPILER, "==", CLANG_CUDA):
        # disable all clang versions older than 14 as CUDA Compiler
        if row_check_version(row, DEVICE_COMPILER, "<", "14"):
            reason(
                output, "all clang versions older than 14 are disabled as CUDA Compiler"
            )
            return False

    return True



from typing import Dict, List

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from packaging import version as pk_version
from typeguard import typechecked

# This dict contains all software version, which we handle we the filter rules.
# Not every software version results in a case in the filter rules. For example
# we have no case, where a combination with a specific Boost version needs to be
# disabled (date: 26.7.2023). Maybe in future, a Boost specific case will be
# added.
versions: Dict[str, List[str]] = {
    GCC: ["6", "7", "8", "9", "10", "11", "12", "13"],
    CLANG: ["6", "7", "8", "9", "10", "11", "12", "13", "14", "15", "16"],
    NVCC: [
        "10.0",
        "10.1",
        "10.2",
        "11.0",
        "11.1",
        "11.2",
        "11.3",
        "11.4",
        "11.5",
        "11.6",
        "11.7",
        "11.8",
        "12.0",
        "12.1",
        "12.2",
    ],
    HIPCC: ["5.0", "5.1", "5.2", "5.3", "5.4", "5.5"],
    ICPX: ["2023.1.0", "2023.2.0"],
    UBUNTU: ["18.04", "20.04"],
    CMAKE: ["3.18", "3.19", "3.20", "3.21", "3.22", "3.23", "3.24", "3.25", "3.26"],
    BOOST: [
        "1.74.0",
        "1.75.0",
        "1.76.0",
        "1.77.0",
        "1.78.0",
        "1.79.0",
        "1.80.0",
        "1.81.0",
        "1.82.0",
    ],
    CXX_STANDARD: ["17", "20"],
}


@typechecked
def is_supported_version(name: str, version: str) -> bool:
    """Check if a specific software version is supported by the
    alpaka-job-coverage library.

    Args:
        name (str): Name of the software, e.g. gcc, boost or ubuntu
        version (str): Version of the software.

    Raises:
        ValueError: When the name of the software is not known.

    Returns:
        bool: True if supported otherwise False.
    """
    if (
        name
        not in list(versions.keys())
        + [
            CLANG_CUDA,
        ]
        + BACKENDS_LIST
    ):
        raise ValueError(f"Unknown software name: {name}")

    local_versions = versions.copy()

    local_versions[CLANG_CUDA] = versions[CLANG]
    local_versions[ALPAKA_ACC_GPU_CUDA_ENABLE] = [OFF_VER] + versions[NVCC]
    local_versions[ALPAKA_ACC_GPU_HIP_ENABLE] = [OFF_VER] + versions[HIPCC]

    for backend_name in set(BACKENDS_LIST) - set(
        (ALPAKA_ACC_GPU_CUDA_ENABLE, ALPAKA_ACC_GPU_HIP_ENABLE)
    ):
        local_versions[backend_name] = [ON_VER, OFF_VER]

    for v in local_versions[name]:
        if pk_version.parse(v) == pk_version.parse(version):
            return True

    return False



"""API for creating sparse job list from parameters list.
"""

import random
from typing import Dict, List, Tuple, Callable

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_job_coverage.filter_compiler_name import general_compiler_filter
from alpaka_job_coverage.filter_compiler_version import compiler_version_filter
from alpaka_job_coverage.filter_backend_version import compiler_backend_filter
from alpaka_job_coverage.filter_software_dependency import (
    software_dependency_filter,
)

from typeguard import typechecked

from allpairspy import AllPairs


@typechecked
def create_job_list(  # pylint: disable=too-many-arguments
    parameters: Dict[str, List],
    pair_size: int = 2,
    pre_filter: Callable[[List], bool] = lambda row: True,
    after_gerneral_compiler_filter: Callable[[List], bool] = lambda row: True,
    after_compiler_version_filter: Callable[[List], bool] = lambda row: True,
    post_filter: Callable[[List], bool] = lambda row: True,
    after_compiler_backend_filter: Callable[[List], bool] = lambda row: True,
) -> List[Dict[str, Tuple[str, str]]]:
    """Create sparse job list from parameters list, respecting the different filter lists.

    Order of the filter functions:
        pre_filter()
        general_compiler_filter()
        after_gerneral_compiler_filter()
        compiler_version_filter()
        after_compiler_version_filter()
        compiler_backend_filter()
        post_filter()

    Args:
        parameters (Dict[str, List]): Dict of parameters with List of values. The keys
        "host_compiler", "device_compiler" and "backends" are required. Each element in the list of
        a dict items needs to be a tuple[str, str]. The first value is the name and second value is
        the version. Only "backends" is an exception. The type of the value of a "backends" item is
        a List[Tuple[str, str]]. Each tuple contains a backend name and version.

        pair_size (int, optional): Each combination of the values of n different parameter fields
        needs to part of at least one job. If there are more than n fields in parameter, it uses
        each possible n combination of the field name. For example, if n is 3 and there are 4
        fields, it will use the following combinations (the numbers represents the field names):
        (0,1,2), (0,1,3), (0,2,3), ... (2,3,4). Than it take the values of each filed of the
        combination and create the full combination matrix. Each of
        this value combinations needs be part of a job. Defaults is 2.

        pre_filter (Callable[[List], bool], optional): This filter is executed first.
        Defaults is a lambda which returns true.

        after_gerneral_compiler_filter (Callable[[List], bool], optional): This
        filter is executed after the general_compiler_filter. Defaults is a lambda which returns
        true.

        after_compiler_version_filter (Callable[[List], bool], optional): This filter
        is executed after the general_compiler_filter. Defaults is a lambda which returns true.

        post_filter (Callable[[List], bool], optional): This filter is executed last.
        Defaults is a lambda which returns true.

    Returns:
        List[Dict[str, List[Tuple[str, str]]]]: Return sparse job matrix.
    """

    # Fill up the param_map.
    # For documentation see the variable documentation in globals.py.
    for index, key in enumerate(parameters):
        param_map[key] = index

    # build filter chain
    filter_function = (
        lambda row: pre_filter(row)
        and general_compiler_filter(row)
        and after_gerneral_compiler_filter(row)
        and compiler_version_filter(row)
        and after_compiler_version_filter(row)
        and compiler_backend_filter(row)
        and after_compiler_backend_filter(row)
        and software_dependency_filter(row)
        and post_filter(row)
    )

    # generate sparse test matrix
    cover_matrix = AllPairs(
        parameters=parameters, n=pair_size, filter_func=filter_function
    )

    # transform the sparse test matrix in a flat data struct
    normalized_cover_matrix: List[Dict[str, Tuple[str, str]]] = []
    for row in cover_matrix:
        new_row = {}
        for index, parameter_name in enumerate(parameters.keys()):
            if parameter_name != BACKENDS:
                new_row[parameter_name] = row[index]
            else:
                # Remove the column backend with its list of alpaka backends and insert each alpaka
                # backend in a own column.
                for backend in row[index]:
                    new_row[backend[NAME]] = backend
        normalized_cover_matrix.append(new_row)

    return normalized_cover_matrix


@typechecked
def shuffle_job_matrix(job_matrix: List[Dict[str, Tuple[str, str]]], seed: int = 42):
    """Shuffle ordering of the job_matrix.

    Args:
        job_matrix (List[Dict[str, Tuple[str, str]]]): The job_matrix.
        seed (int, optional): Seed of the shuffle function. Defaults to 42.
    """
    random.Random(seed).shuffle(job_matrix)


@typechecked
def distribute_to_waves(
    job_matrix: List[Dict[str, Dict]], wave_size: int, max_jobs: int = 0
) -> List[List[Dict[str, Dict]]]:
    """Distribute the jobs of job_matrix (1D) to a list of waves (2D). Each wave has the size
    wave_size. Total number in list waves is smaller or equal to max_jobs.

    Args:
        job_matrix (List[Dict[str, Tuple[str, str]]]): The job_matrix.
        wave_size (int): Size of each wave.
        max_jobs (int, optional): Maximum number of jobs in the returned matrix. Defaults to 0.

    Returns:
        List[List[Dict[str, Tuple[str, str]]]]: 2D Matrix with wave_size jobs each outer list.
    """
    # if max_jobs is 0, set to len(job_matrix)
    max_jobs = max_jobs if max_jobs > 0 else len(job_matrix)

    # if max_jobs is greater than len(job_matrix), crop it to len(job_matrix)
    max_jobs = max_jobs if max_jobs < len(job_matrix) else len(job_matrix)

    wave_matrix: List[List[Dict[str, Dict]]] = []

    for i in range(0, max_jobs, wave_size):
        wave_matrix.append(job_matrix[i : i + wave_size])

    # if wave_size is not a multiple of max_jobs and max_jobs is smaller than len(job_matrix),
    # the last wave contains to much jobs
    # this statement removes the jobs that are too much
    if max_jobs % wave_size != 0 and len(wave_matrix[-1]) > max_jobs % wave_size:
        wave_matrix[-1] = wave_matrix[-1][: max_jobs % wave_size]

    return wave_matrix



import unittest


from alpaka_job_coverage.versions import is_supported_version
from alpaka_job_coverage.globals import *
from utils import manual_version_test


class TestSupportedVersion(unittest.TestCase):
    def test_name(self):
        try:
            for name in [
                GCC,
                CLANG,
                NVCC,
                CLANG_CUDA,
                HIPCC,
                ALPAKA_ACC_GPU_CUDA_ENABLE,
                ALPAKA_ACC_GPU_HIP_ENABLE,
            ]:
                is_supported_version(name, "0")
        except ValueError as e:
            self.fail(f'"{name}" should be accepted by is_supported_version()')

        # non_existing_compiler should fail, because the compiler does not exist
        non_existing_compiler = "magic_compiler"
        self.assertRaises(
            ValueError,
            is_supported_version,
            non_existing_compiler,
            "0",
        )

    def test_version(self):
        test_versions: Dict[str, Tuple[List[str], List[str]]] = {
            GCC: (["7", "9", "10"], ["99"]),
            CLANG: (["8", "10", "14"], ["99", "365"]),
            NVCC: (["10.1", "11.4", "12.1"], ["99.9"]),
            HIPCC: (["5.0", "5.3", "5.5"], ["99.9"]),
            UBUNTU: (["18.04", "20.04"], ["99.04", "2004.04"]),
        }

        # manual check of the version
        # the tests only works, if we use the following supported and
        # unsupported versions
        for name, (supported_versions, unsupported_versions) in test_versions.items():
            manual_version_test(self, name, supported_versions, unsupported_versions)

        for name, (supported_versions, unsupported_versions) in test_versions.items():
            for v in supported_versions:
                self.assertTrue(
                    is_supported_version(name, v), f"{name} {v} should be supported."
                )

                # we support the same clang versions also for clang as CUDA compiler
                if name == CLANG:
                    self.assertTrue(
                        is_supported_version(CLANG_CUDA, v),
                        f"{CLANG_CUDA} {v} should be supported.",
                    )

                # the nvcc compiler is bundled with the CUDA SDK and both
                # has the same version
                if name == NVCC:
                    self.assertTrue(
                        is_supported_version(ALPAKA_ACC_GPU_CUDA_ENABLE, v),
                        f"{ALPAKA_ACC_GPU_CUDA_ENABLE} {v} should be supported.",
                    )

                # the hipcc compiler is bundled with the ROCm SDK and both
                # has the same version
                if name == HIPCC:
                    self.assertTrue(
                        is_supported_version(ALPAKA_ACC_GPU_HIP_ENABLE, v),
                        f"{ALPAKA_ACC_GPU_HIP_ENABLE} {v} should be supported.",
                    )

            for v in unsupported_versions:
                self.assertFalse(
                    is_supported_version(name, v),
                    f"{name} {v} should be not supported.",
                )

                # we support the same clang versions also for clang as CUDA compiler
                if name == CLANG:
                    self.assertFalse(
                        is_supported_version(CLANG_CUDA, v),
                        f"{CLANG_CUDA} {v} should be not supported.",
                    )

                # the nvcc compiler is bundled with the CUDA SDK and both
                # has the same version
                if name == NVCC:
                    self.assertFalse(
                        is_supported_version(ALPAKA_ACC_GPU_CUDA_ENABLE, v),
                        f"{ALPAKA_ACC_GPU_CUDA_ENABLE} {v} should be not supported.",
                    )

                # the hipcc compiler is bundled with the ROCm SDK and both
                # has the same version
                if name == HIPCC:
                    self.assertFalse(
                        is_supported_version(ALPAKA_ACC_GPU_HIP_ENABLE, v),
                        f"{ALPAKA_ACC_GPU_HIP_ENABLE} {v} should be not supported.",
                    )

    def test_on_off_backends(self):
        # verify that

        # NVCC and ALPAKA_ACC_GPU_CUDA_ENABLE shares the same versions
        manual_version_test(self, NVCC, [], [ON_VER, "2.7", "3.0"])
        # HIPCC and ALPAKA_ACC_GPU_HIP_ENABLE shares the same versions
        manual_version_test(self, HIPCC, [], [ON_VER, "2.7", "3.0"])

        for backend_name in BACKENDS_LIST:
            if (
                backend_name != ALPAKA_ACC_GPU_CUDA_ENABLE
                and backend_name != ALPAKA_ACC_GPU_HIP_ENABLE
            ):
                for correct_version in [ON_VER, OFF_VER]:
                    self.assertTrue(
                        is_supported_version(backend_name, correct_version),
                        f"{backend_name} {correct_version} should be allowed.",
                    )
            else:
                self.assertTrue(
                    is_supported_version(backend_name, OFF_VER),
                    f"{backend_name} {correct_version} should be allowed.",
                )
                self.assertFalse(
                    is_supported_version(backend_name, ON_VER),
                    f"{backend_name} {correct_version} should be allowed.",
                )

            for correct_version in [
                "2.7",
                "3.0",
            ]:
                self.assertFalse(
                    is_supported_version(backend_name, correct_version),
                    f"{backend_name} {correct_version} is not supported.",
                )



"""This module contains constances used in the alpaka job coverage framework.
"""

from typing import List, Dict

# index positions of the name version tuple, used in the paramer list
NAME: int = 0
VERSION: int = 1

# parameter key names, whit special meaning
HOST_COMPILER: str = "host_compiler"
DEVICE_COMPILER: str = "device_compiler"
BACKENDS: str = "backends"

# name of the used compilers
GCC: str = "gcc"
CLANG: str = "clang"
NVCC: str = "nvcc"
CLANG_CUDA: str = "clang-cuda"
HIPCC: str = "hipcc"
ICPX: str = "icpx"

# alpaka backend names
ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE: str = "alpaka_ACC_CPU_B_SEQ_T_SEQ_ENABLE"
ALPAKA_ACC_CPU_B_SEQ_T_THREADS_ENABLE: str = "alpaka_ACC_CPU_B_SEQ_T_THREADS_ENABLE"
ALPAKA_ACC_CPU_B_SEQ_T_FIBERS_ENABLE: str = "alpaka_ACC_CPU_B_SEQ_T_FIBERS_ENABLE"
ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE: str = "alpaka_ACC_CPU_B_TBB_T_SEQ_ENABLE"
ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE: str = "alpaka_ACC_CPU_B_OMP2_T_SEQ_ENABLE"
ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE: str = "alpaka_ACC_CPU_B_SEQ_T_OMP2_ENABLE"
ALPAKA_ACC_ANY_BT_OMP5_ENABLE: str = "alpaka_ACC_ANY_BT_OMP5_ENABLE"
ALPAKA_ACC_GPU_CUDA_ENABLE: str = "alpaka_ACC_GPU_CUDA_ENABLE"
ALPAKA_ACC_GPU_HIP_ENABLE: str = "alpaka_ACC_GPU_HIP_ENABLE"
ALPAKA_ACC_SYCL_ENABLE: str = "alpaka_ACC_SYCL_ENABLE"
BACKENDS_LIST: List[str] = [
    ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE,
    ALPAKA_ACC_CPU_B_SEQ_T_THREADS_ENABLE,
    ALPAKA_ACC_CPU_B_SEQ_T_FIBERS_ENABLE,
    ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE,
    ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE,
    ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE,
    ALPAKA_ACC_ANY_BT_OMP5_ENABLE,
    ALPAKA_ACC_GPU_CUDA_ENABLE,
    ALPAKA_ACC_GPU_HIP_ENABLE,
    ALPAKA_ACC_SYCL_ENABLE,
]

# if want to use version comparison functions from util.py, you need to use ON_VER and OFF_VER
ON: str = "on"
OFF: str = "off"

# backend states
ON_VER: str = "1.0.0"
OFF_VER: str = "0.0.0"

# additional parameters, like alpaka software dependencies, compiler configurations and so one
UBUNTU: str = "ubuntu"
CMAKE: str = "cmake"
BOOST: str = "boost"
CXX_STANDARD: str = "cxx_standard"

# Is required because allpairspy uses a List for each row. The ordering of the entries in a row
# is the same, like the order of the keys of the parameters dict. To make the filter function
# independent of the number and ordering of the entries in parameters, the param_map is
# required.
#
# the map is filled up by the main_functions.create_job_list() function
param_map: Dict[str, int] = {}



#!/usr/bin/env python3

import argparse
from packaging import version as pk_version
from typing import Dict, Tuple, Union
from typeguard import typechecked
from io import StringIO

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
import alpaka_job_coverage.filter_compiler_name as ajc_compiler_name
import alpaka_job_coverage.filter_compiler_version as ajc_compiler_version
import alpaka_job_coverage.filter_backend_version as ajc_backend_version
import alpaka_job_coverage.filter_software_dependency as ajc_software_dependency
from alpaka_job_coverage.versions import is_supported_version


@typechecked
def cs(text: str, color: str):
    """Prints colored text to the command line. The text printed after
       the function call has the default color of the command line.

    :param text: text to be colored
    :type text: str
    :param color: Name of the color. If color is unknown or empty use default color
                  of the command line.
    :type color: str
    :returns:
    :rtype:

    """
    if color is None:
        return text

    output = ""
    if color == "Red":
        output += "\033[0;31m"
    elif color == "Green":
        output += "\033[0;32m"
    elif color == "Yellow":
        output += "\033[1;33m"
    else:
        return text

    return output + text + "\033[0m"


@typechecked
def exit_error(text: str):
    """Prints error message and exits application with error code 1.

    Args:
        text (str): Error message.
    """
    print(cs("ERROR: " + text, "Red"))
    exit(1)


@typechecked
def validate_args_compiler(arguments: Dict[str, str]) -> Dict[str, Tuple[str, str]]:
    """Checks if the compiler name exists and the version number is parsable.

    Args:
        arguments (Dict[str, str]): Key needs to be HOST_COMPILER or
        DEVICE_COMPILER and values describes the compiler name and version.

    Returns:
        Dict[str, Tuple[str, str]]: Returns parsed compiler name and version
    """
    validated_args = {}
    known_compiler = [GCC, CLANG, NVCC, CLANG_CUDA, HIPCC, ICPX]

    for parameter_name, parameter in arguments.items():
        if "@" not in parameter:
            exit_error(f"@ is missing in {parameter_name}={parameter}")

        splitted_name_version = parameter.split("@", 1)
        name = splitted_name_version[0]
        version = splitted_name_version[1]

        if name not in known_compiler:
            exit_error(f"Unknown compiler: {name}\nKnown compilers: {known_compiler}")

        # use parse() function to validate that the version has a valid shape
        try:
            pk_version.parse(version)
        except Exception as e:
            exit_error(f"Could not parse version number of {name}: {version}")

        validated_args[parameter_name] = (name, version)

    return validated_args


@typechecked
def validate_args_backend(arguments: List[str]) -> Dict[str, List[Tuple[str, str]]]:
    """Checks if back-end names exist and the version numbers are parsable.

    Args:
        arguments (List[str]): List of back-end names with versions.

    Returns:
        Dict[str, List[Tuple[str, str]]]: List of parsed back-end names with
        versions.
    """
    validated_args = {BACKENDS: []}

    for backend in arguments:
        if backend not in BACKENDS_LIST:
            if "@" not in backend:
                exit_error(f"@ is missing in {backend}")

            splitted_name_version = backend.split("@", 1)
            name = splitted_name_version[0]
            version = splitted_name_version[1]

            if name not in BACKENDS_LIST:
                exit_error(f"Unknown back-end: {name}\nKnown back-ends: {BACKENDS_LIST}")

            # use parse() function to validate that the version has a valid shape
            try:
                pk_version.parse(version)
            except Exception as e:
                exit_error(f"Could not parse version number of {name}: {version}")

            validated_args[BACKENDS].append((name, version))

    return validated_args


@typechecked
class VersionAction(argparse.Action):
    # check if argument has valid version shape
    def __call__(self, parser, namespace, values, option_string):
        try:
            pk_version.parse(values)
        except Exception as e:
            print(cs(f"ERROR: wrong version for {option_string}", "Red"))
            raise e
        setattr(namespace, self.dest, values)


@typechecked
def validate_version_support(
    parameters: Dict[str, Union[Tuple[str, str], List[Tuple[str, str]]]],
):
    """Prints a warning if a software version is not officially supported by the
    library. This function does not exit the application because the versions
    can still be tested with the filter functions.

    Args:
        parameters (Dict[str, Union[Tuple[str, str], List[Tuple[str, str]]]]):
        parameter set of the software versions to test
    """
    for param_name, param_value in parameters.items():
        if param_name == BACKENDS:
            for backend_name, backend_version in param_value:
                if not is_supported_version(backend_name, backend_version):
                    print(
                        cs(
                            f"WARNING: {backend_name} {backend_version} is not officially supported.",
                            "Yellow",
                        )
                    )
        else:
            if not is_supported_version(param_value[0], param_value[1]):
                print(
                    cs(
                        f"WARNING: {param_value[0]} {param_value[1]} is not officially supported.",
                        "Yellow",
                    )
                )


@typechecked
def check_single_filter(
    filter: callable,
    req_params: List[str],
    parameters: Dict[str, Union[Tuple[str, str], List[Tuple[str, str]]]],
) -> bool:
    """Applies a parameter set on the filter function if it fulfils all requirements
    and returns the result. The parameter set needs to contain all parameters which
    are defined in `req_params`.

    Args:
        filter (callable): filter function
        req_params (List[str]): list of required parameter names
        parameters (Dict[str, Union[Tuple[str, str], List[Tuple[str, str]]]]):
        parameter set

    Returns:
        bool: returns True if parameter set passes the filter
    """
    missing_parameters = ""

    # get name of the filter for command line output.
    filter_name: str = filter.__name__

    # Each filter function has also type checked version ending with _typed.
    # Remove _typed for better output.
    if filter_name.endswith("_typed"):
        filter_name = filter_name[: -len("_typed")]

    # check if all required parameter are available
    for req_name in req_params:
        if req_name not in parameters.keys():
            missing_parameters += " " + req_name

    if missing_parameters:
        print(
            cs(
                f"skipped {filter_name}(), missing parameters ->" + missing_parameters,
                "Yellow",
            )
        )
        return False
    else:
        global param_map
        row: List[Union[Tuple[str, str], List[Tuple[str, str]]]] = []

        # convert parameter map to list and configure param_map for the
        # lookup
        for i, (param_name, param) in enumerate(parameters.items()):
            param_map[param_name] = i
            row.append(param)

        # the msg object allows to get information from filter function, why
        # parameter combination does not pass the filter
        msg = StringIO()

        if filter(row, msg):
            print(cs(f"{filter_name}() returns True", "Green"))
            # reset param_map
            param_map = {}
            return True
        else:
            print(cs(f"{filter_name}() returns False", "Red"))
            if msg.getvalue() != "":
                print("  " + msg.getvalue())
            # reset param_map
            param_map = {}
            return False


@typechecked
def check_filters(
    parameters: Dict[str, Union[Tuple[str, str], List[Tuple[str, str]]]]
) -> bool:
    """Applies a parameter set to all available filters in the same order, in the same way
    the ajc library is doing it.

    Args:
        parameters (Dict[str, Union[Tuple[str, str], List[Tuple[str, str]]]]):
        parameter set

    Returns:
        bool: returns True if all filters return True
    """
    all_true = 0
    all_true += int(
        check_single_filter(
            ajc_compiler_name.general_compiler_filter_typed,
            ajc_compiler_name.get_required_parameter(),
            parameters,
        )
    )
    all_true += int(
        check_single_filter(
            ajc_compiler_version.compiler_version_filter_typed,
            ajc_compiler_version.get_required_parameter(),
            parameters,
        )
    )
    all_true += int(
        check_single_filter(
            ajc_backend_version.compiler_backend_filter_typed,
            ajc_backend_version.get_required_parameter(),
            parameters,
        )
    )
    all_true += int(
        check_single_filter(
            ajc_software_dependency.software_dependency_filter_typed,
            ajc_software_dependency.get_required_parameter(),
            parameters,
        )
    )

    # each filter add a one, if it was successful
    return all_true == 4


def main():
    parser = argparse.ArgumentParser(
        description="Check if combination of parameters is valid."
    )

    parser.add_argument(
        "--host-compiler",
        type=str,
        help="Define host compiler. Shape needs to be name@version. "
        "For example gcc@10",
    )

    parser.add_argument(
        "--device-compiler",
        type=str,
        help="Define device compiler. Shape needs to be name@version. "
        "For example nvcc@11.3",
    )

    parser.add_argument(
        "--backends",
        nargs="*",
        help="Define back-ends as whitespace separated list. Each element needs "
        "to have the shape of name@version. For example: "
        "alpaka_ACC_CPU_B_SEQ_T_SEQ_ENABLE@1 "
        "ALPAKA_ACC_CPU_B_SEQ_T_THREADS_ENABLE@0 "
        "ALPAKA_ACC_GPU_CUDA_ENABLE@11.3\n"
        "Use the values 0 and 1 to disable or enable a back-end.",
    )

    parser.add_argument(
        "--ubuntu", type=str, choices=["18.04", "20.04"], help="Ubuntu version."
    )

    parser.add_argument(
        "--cmake", type=str, action=VersionAction, help="Set CMake version."
    )

    parser.add_argument(
        "--boost", type=str, action=VersionAction, help="Set Boost version."
    )

    parser.add_argument("--cxx", type=str, choices=["17", "20"], help="C++ version.")

    parser.add_argument(
        "--print-backends", action="store_true", help="Print all available back-ends."
    )

    parser.add_argument(
        "--print-parameters", action="store_true", help="Print validated parameters."
    )

    args = parser.parse_args()

    if args.print_backends:
        for b in BACKENDS_LIST:
            print(b)
        exit(0)

    parameters = {}

    ############################################################################
    ## parse compiler
    ############################################################################

    if args.host_compiler:
        parameters.update(validate_args_compiler({HOST_COMPILER: args.host_compiler}))

    if args.device_compiler:
        parameters.update(
            validate_args_compiler({DEVICE_COMPILER: args.device_compiler})
        )

    if args.backends:
        parameters.update(validate_args_backend(args.backends))

    if args.ubuntu:
        parameters[UBUNTU] = (UBUNTU, args.ubuntu)

    if args.cmake:
        parameters[CMAKE] = (CMAKE, args.cmake)

    if args.boost:
        parameters[BOOST] = (BOOST, args.boost)

    if args.cxx:
        parameters[CXX_STANDARD] = (CXX_STANDARD, args.cxx)

    if args.print_parameters:
        print(parameters)

    validate_version_support(parameters)
    exit(int(not check_filters(parameters)))


if __name__ == "__main__":
    main()



"""Filter rules basing on backend names and versions.
"""

import io

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_job_coverage.util import (
    row_check_name,
    row_check_version,
    row_check_backend_version,
    reason,
)
from typing import List, Tuple, Union, Optional
from typeguard import typechecked


def get_required_parameter() -> List[str]:
    """Returns a list of parameters which are required for using the filter defined by this module.

    Returns:
        List[str]: list of parameters
    """
    return [DEVICE_COMPILER, BACKENDS]


@typechecked
def compiler_backend_filter_typed(
    row: List[Union[Tuple[str, str], List[Tuple[str, str]]]],
    output: Optional[Union[io.StringIO, io.TextIOWrapper]] = None,
) -> bool:
    """Type checked version of compiler_backend_filter(). Should be only used for
    testing or tooling. The type check adds a big overhead, which slows down
    pair-wise generator by the factor 30.

    Args:
        row (List[Union[Tuple[str, str], List[Tuple[str, str]]]]): Combination
        to verify. The row can contain up to all combination fields and at least
        two items.
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): Write
        additional information about filter decisions to the IO object
        (io.SringIO, sys.stdout, sys.stderr). If it is None no information is
        generated.

    Returns:
        bool: True, if combination is valid, otherwise False.
    """
    return compiler_backend_filter(row, output)


def compiler_backend_filter(
    row: List, output: Optional[Union[io.StringIO, io.TextIOWrapper]] = None
) -> bool:
    """Filter rules basing on backend names and versions.

    Args:
        row (List): Combination to verify. The row can contain
        up to all combination fields and at least two items.
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): Write
        additional information about filter decisions to the IO object
        (io.SringIO, sys.stdout, sys.stderr). If it is None no information is
        generated.

    Returns:
        bool: True, if combination is valid, otherwise False.
    """
    ###########################
    ## gcc device compiler
    ###########################

    if row_check_name(row, DEVICE_COMPILER, "==", GCC):
        if row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "gcc as device compiler cannot compile with enabled "
                "alpaka_ACC_GPU_CUDA_ENABLE back-end",
            )
            return False

        if row_check_backend_version(row, ALPAKA_ACC_GPU_HIP_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "gcc as device compiler cannot compile with enabled "
                "alpaka_ACC_GPU_HIP_ENABLE back-end",
            )
            return False
        
        if row_check_backend_version(row, ALPAKA_ACC_SYCL_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "gcc as device compiler cannot compile with enabled "
                "alpaka_ACC_SYCL_ENABLE back-end",
            )
            return False

    ###########################
    ## clang device compiler
    ###########################

    if row_check_name(row, DEVICE_COMPILER, "==", CLANG):
        if row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "clang as device compiler cannot compile with enabled "
                "alpaka_ACC_GPU_CUDA_ENABLE back-end (use clang-cuda instead)",
            )
            return False

        # clang cannot compile with enabled HIP backend
        if row_check_backend_version(row, ALPAKA_ACC_GPU_HIP_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "clang as device compiler cannot compile with enabled "
                "alpaka_ACC_GPU_HIP_ENABLE back-end (use hipcc instead)",
            )
            return False
        
        if row_check_backend_version(row, ALPAKA_ACC_SYCL_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "clang as device compiler cannot compile with enabled "
                "alpaka_ACC_SYCL_ENABLE back-end (use icpx instead)",
            )
            return False

    ###########################
    ## nvcc device compiler
    ###########################

    if row_check_name(row, DEVICE_COMPILER, "==", NVCC):       
        # the nvcc compiler needs the same version, like the backend
        if row_check_backend_version(
            row,
            ALPAKA_ACC_GPU_CUDA_ENABLE,
            "!=",
            row[param_map[DEVICE_COMPILER]][VERSION],
        ):
            reason(
                output,
                "the nvcc compiler and the CUDA back-end must have the same version",
            )
            return False

        # it is not allowed to enable the HIP and CUDA backend on the same time
        if row_check_backend_version(row, ALPAKA_ACC_GPU_HIP_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "If nvcc is the device compiler and the CUDA back-end is enabled "
                "it is not allowed to enable the HIP back-end",
            )
            return False
        
        if row_check_backend_version(row, ALPAKA_ACC_SYCL_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "If nvcc is the device compiler it is not allowed to enable "
                "the SYCL back-end"
            )
            return False

    ###########################
    ## clang-cuda device compiler
    ###########################

    if row_check_name(row, DEVICE_COMPILER, "==", CLANG_CUDA):
        # the CUDA backend needs to be enabled
        if row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, "==", OFF_VER):
            reason(
                output,
                "when CLANG_CUDA is set as device compiler "
                "the CUDA back-end must be enabled",
            )
            return False

        # it is not allowed to enable the HIP and CUDA backend on the same time
        if row_check_backend_version(row, ALPAKA_ACC_GPU_HIP_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "if clang-cuda is the device compiler and the CUDA back-end is enabled "
                "it is not allowed to enable the HIP back-end",
            )
            return False
        
        # clang-cuda doesn't support the SYCL back-end
        if row_check_backend_version(row, ALPAKA_ACC_SYCL_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "if clang-cuda is the device compiler it is not allowed to "
                "enable the SYCL back-end",
            )
            return False

        # check if clang-cuda supports the CUDA SDK version
        clangcuda_cudasdk_versions = [
            ("7", "9.2"),
            ("8", "10.0"),
            ("10", "10.1"),
            ("12", "11.0"),
            ("13", "11.2"),
            ("16", "11.5"),
        ]

        for clang_cuda_version, cuda_sdk_version in clangcuda_cudasdk_versions:
            if row_check_version(
                row, DEVICE_COMPILER, "<=", clang_cuda_version
            ) and row_check_backend_version(
                row, ALPAKA_ACC_GPU_CUDA_ENABLE, ">", cuda_sdk_version
            ):
                reason(
                    output,
                    f"clang-{clang_cuda_version} supports only up to CUDA {cuda_sdk_version}",
                )
                return False

    ###########################
    ## hipcc device compiler
    ###########################

    # the HIP backend needs to be enabled and has the same version number
    if row_check_name(row, DEVICE_COMPILER, "==", HIPCC):
        if row_check_backend_version(
            row,
            ALPAKA_ACC_GPU_HIP_ENABLE,
            "!=",
            row[param_map[DEVICE_COMPILER]][VERSION],
        ):
            reason(
                output, "hipcc and the HIP back-end must have the same version number"
            )
            return False

        # it is not allowed to enable the HIP and CUDA backend on the same time
        if row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "if hipcc is the device compiler and the HIP back-end is enabled, "
                "it is not allowed to enable the CUDA back-end",
            )
            return False
        
        if row_check_backend_version(row, ALPAKA_ACC_SYCL_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "if hipcc is the device compiler it is not allowed to enable "
                "the SYCL back-end",
            )
            return False
        
    ###########################
    ## icpx device compiler
    ###########################

    # Don't use icpx for the CUDA and HIP back-ends
    if row_check_name(row, DEVICE_COMPILER, "==", ICPX):
        if row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "if icpx is the device compiler it is not allowed to enable the CUDA "
                "back-end",
            )
            return False
        
        if row_check_backend_version(row, ALPAKA_ACC_GPU_HIP_ENABLE, "!=", OFF_VER):
            reason(
                output,
                "if icpx is the device compiler it is not allowed to enable the HIP "
                "back-end",
            )
            return False
    return True



import unittest

from alpaka_job_coverage.filter_compiler_name import general_compiler_filter_typed
from alpaka_job_coverage.filter_compiler_version import compiler_version_filter_typed
from alpaka_job_coverage.filter_backend_version import compiler_backend_filter_typed
from alpaka_job_coverage.filter_software_dependency import (
    software_dependency_filter_typed,
)
from alpaka_job_coverage.globals import *
from utils import manual_version_test


def full_filter_chain(row) -> bool:
    return (
        general_compiler_filter_typed(row)
        and compiler_version_filter_typed(row)
        and compiler_backend_filter_typed(row)
        and software_dependency_filter_typed(row)
    )

class TestIcpxHostDeviceCompilerBackend(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[HOST_COMPILER] = 0
        param_map[DEVICE_COMPILER] = 1
        param_map[BACKENDS] = 2

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    # test that no combination is forbidden when only icpx is the host compiler
    # and no other parameter is set
    def test_only_host_compiler(self):
        valid_combs = [
            [(ICPX, "0")],
            [(ICPX, "2023.1.0")],
            [(ICPX, "2023.2.0")],
        ]

        for comb in valid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}",
            )
            self.assertTrue(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]}",
            )

    # verify that only icpx can be used if it is set as host and device compiler
    # (ignore compiler version)
    def test_host_device_compiler_name(self):
        valid_combs = [
            [(ICPX, "0"), (ICPX, "0")],
        ]

        for comb in valid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )

        invalid_combs = [
            # CLANG_CUDA as device compiler is not tested, because a rule
            # forbids to use clang-cuda < 14
            [(ICPX, "0"), (NVCC, "0")],
            [(ICPX, "0"), (GCC, "0")],
            [(ICPX, "0"), (CLANG, "0")],
            [(NVCC, "0"), (ICPX, "0")],
            [(GCC, "0"), (ICPX, "0")],
            [(CLANG, "0"), (ICPX, "0")],
            [(CLANG_CUDA, "0"), (ICPX, "0")],
        ]

        for comb in invalid_combs:
            self.assertFalse(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertFalse(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )

    # verify that only icpx can be used, if it set as host and device compiler
    # and has the same compiler version
    def test_host_device_compiler_version(self):
        valid_combs = [
            [(ICPX, "2023.1.0"), (ICPX, "2023.1.0")],
            [(ICPX, "2023.2.0"), (ICPX, "2023.2.0")],
        ]

        for comb in valid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )

        invalid_combs = [
            [(ICPX, "2023.1.0"), (ICPX, "2023.2.0")],
            [(ICPX, "2023.2.0"), (ICPX, "2023.1.0")],
        ]

        for comb in invalid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertFalse(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertFalse(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )

    # Allowed back-ends are SYCL and the CPU back-ends
    def test_host_device_compiler_backend(self):
        valid_combs = [
            [(ICPX, "2023.1.0"), (ICPX, "2023.1.0"), [(ALPAKA_ACC_SYCL_ENABLE, ON_VER)]],
            [(ICPX, "2023.2.0"), (ICPX, "2023.2.0"), [(ALPAKA_ACC_SYCL_ENABLE, ON_VER)]],
            [
                (ICPX, "2023.1.0"),
                (ICPX, "2023.1.0"),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_SYCL_ENABLE, ON_VER),
                ],
            ],
            [
                (ICPX, "2023.2.0"),
                (ICPX, "2023.2.0"),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_GPU_HIP_ENABLE, OFF_VER),
                    (ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER),
                ],
            ],
        ]

        for comb in valid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )

        invalid_combs = [
            [(ICPX, "2023.1.0"), (ICPX, "2023.1.0"), [(ALPAKA_ACC_GPU_HIP_ENABLE, "4.4")]],
            [(ICPX, "2023.2.0"), (ICPX, "2023.2.0"), [(ALPAKA_ACC_GPU_HIP_ENABLE, "5.2")]],
            [(ICPX, "2023.1.0"), (ICPX, "2023.1.0"), [(ALPAKA_ACC_GPU_CUDA_ENABLE, "12.1")]],
            [
                (ICPX, "2023.2.0"),
                (ICPX, "2023.2.0"),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_GPU_HIP_ENABLE, "5.5"),
                ],
            ],
            # it is forbidden to enable the CUDA and SYCL back-end at the same time
            [
                (ICPX, "2023.1.0"),
                (ICPX, "2023.1.0"),
                [
                    (ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2"),
                    (ALPAKA_ACC_SYCL_ENABLE, ON_VER),
                ],
            ],
        ]

        for comb in invalid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertFalse(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertFalse(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )


"""Filter rules handling software dependencies and compiler settings.
"""

import io

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_job_coverage.util import (
    row_check_backend_version,
    row_check_name,
    row_check_version,
    is_in_row,
    reason,
)

from packaging import version as pk_version
from typing import List, Tuple, Union, Optional
from typeguard import typechecked


def get_required_parameter() -> List[str]:
    """Returns a list of parameters which are required for using the filter defined by this module.
    Returns:
        List[str]: list of parameters
    """
    return [HOST_COMPILER, DEVICE_COMPILER, BACKENDS, UBUNTU, CMAKE, CXX_STANDARD]


@typechecked
def software_dependency_filter_typed(
    row: List[Union[Tuple[str, str], List[Tuple[str, str]]]],
    output: Optional[Union[io.StringIO, io.TextIOWrapper]] = None,
) -> bool:
    """Type checked version of software_dependency_filter(). Should be only used for
    testing or tooling. The type check adds a big overhead, which slows down
    pair-wise generator by the factor 30.

    Args:
        row (List[Union[Tuple[str, str], List[Tuple[str, str]]]]): Combination
        to verify. The row can contain up to all combination fields and at least
        two items.
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): Write
        additional information about filter decisions to the IO object
        (io.SringIO, sys.stdout, sys.stderr). If it is None no information is
        generated.

    Returns:
        bool: True, if combination is valid, otherwise False.
    """
    return software_dependency_filter(row, output)


def software_dependency_filter(
    row: List, output: Optional[Union[io.StringIO, io.TextIOWrapper]] = None
) -> bool:
    """Filter rules handling software dependencies and compiler settings.

    Args:
        row (List): Combination to verify. The row can contain
        up to all combination fields and at least two items.
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): Write
        additional information about filter decisions to the IO object
        (io.SringIO, sys.stdout, sys.stderr). If it is None, no information is
        generated.

    Returns:
        bool: True, if combination is valid, otherwise False.
    """

    # GCC 6 and below is not available on Ubuntu 20.04
    if row_check_version(row, UBUNTU, "==", "20.04"):
        if (
            row_check_name(row, HOST_COMPILER, "==", GCC)
            and int(row[param_map[HOST_COMPILER]][VERSION]) <= 6
        ):
            reason(output, "gcc versions <= 6 are not available on Ubuntu 20.04")
            return False

    # GCC 9 and older does not support -std=c++20
    if (
        row_check_version(row, CXX_STANDARD, ">=", "20")
        and row_check_name(row, HOST_COMPILER, "==", GCC)
        and row_check_version(row, HOST_COMPILER, "<=", "9")
    ):
        reason(output, "gcc versions <= 9 do not support C++20")
        return False

    if row_check_name(row, DEVICE_COMPILER, "==", NVCC) and is_in_row(
        row, CXX_STANDARD
    ):
        parsed_nvcc_version = pk_version.parse(row[param_map[DEVICE_COMPILER]][VERSION])

        # definition of the tuple values: if the nvcc version of the first
        # tuple is older than the cxx standard of the second value, it is not supported
        nvcc_cxx_versions = [
            ("11.0", 17),  # NVCC versions older than 11.0 does not support C++ 17
            ("12.0", 20),  # NVCC versions older than 12.0 does not support C++ 20
            ("12.3", 23),  # NVCC 12.3 is not released yet, therefore we need to
            # expect that it could support C++23
        ]
        for nvcc_version, cxx_version in nvcc_cxx_versions:
            if (
                parsed_nvcc_version < pk_version.parse(nvcc_version)
                and int(row[param_map[CXX_STANDARD]][VERSION]) >= cxx_version
            ):
                reason(
                    output,
                    f"nvcc-{row[param_map[DEVICE_COMPILER]][VERSION]} "
                    f"does not support C++{cxx_version}",
                )
                return False

    # clang 11 and 12 are not available in the Ubuntu 18.04 ppa
    if (
        row_check_version(row, UBUNTU, "==", "18.04")
        and (
            row_check_name(
                row,
                HOST_COMPILER,
                "==",
                CLANG,
            )
            or row_check_name(
                row,
                HOST_COMPILER,
                "==",
                CLANG_CUDA,
            )
        )
        and (
            (
                row_check_version(row, HOST_COMPILER, "==", "11")
                or row_check_version(row, HOST_COMPILER, "==", "12")
            )
        )
    ):
        reason(output, "clang 11 and 12 are not available in the Ubuntu 18.04 PPA")
        return False

    # Clang 9 and older does not support -std=c++20
    if row_check_version(row, CXX_STANDARD, ">=", "20"):
        for compiler_name in [CLANG, CLANG_CUDA]:
            if row_check_name(
                row, HOST_COMPILER, "==", compiler_name
            ) and row_check_version(row, HOST_COMPILER, "<=", "9"):
                reason(output, "clang versions <= 9 do not support C++20")
                return False

    # ubuntu 18.04 containers are not available for CUDA 11.0 and later
    if (
        row_check_version(row, UBUNTU, "==", "18.04")
        and row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, "!=", OFF_VER)
        and row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, ">=", "11.0")
    ):
        reason(
            output, "Ubuntu 18.04 containers are not available for CUDA 11.0 and later"
        )
        return False

    # ubuntu 20.04 containers are not available for CUDA 10.2 and before
    if (
        row_check_version(row, UBUNTU, "==", "20.04")
        and row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, "!=", OFF_VER)
        and row_check_backend_version(row, ALPAKA_ACC_GPU_CUDA_ENABLE, "<", "11.0")
    ):
        reason(
            output, "Ubuntu 20.04 containers are not available for CUDA 10.2 and before"
        )
        return False

    # all rocm images are Ubuntu 20.04 based
    if (
        row_check_version(row, UBUNTU, "!=", "20.04")
        and row_check_name(row, DEVICE_COMPILER, "==", HIPCC)
        and row_check_backend_version(row, ALPAKA_ACC_GPU_HIP_ENABLE, "!=", OFF_VER)
    ):
        reason(output, "all ROCm images are based on Ubuntu 20.04")
        return False

    # a bug in CMAKE 3.18 avoids the correct usage of the variable CMAKE_CUDA_ARCHITECTURE if the
    # CUDA compiler is Clang++
    if row_check_name(row, DEVICE_COMPILER, "==", CLANG_CUDA) and row_check_version(
        row, CMAKE, "<", "3.19"
    ):
        reason(
            output,
            "CMake versions <= 3.18 do not support clang-cuda correctly.",
        )
        return False

    # disable nvcc 11.0-11.3 + gcc 10 + Ubuntu 20.04
    # Ubuntu 20.04 provides only gcc 10.3 and not 10.4 or 10.5
    # this combination does not work: https://github.com/alpaka-group/alpaka/issues/1297
    if (
        row_check_name(row, DEVICE_COMPILER, "==", NVCC)
        and (
            row_check_version(row, DEVICE_COMPILER, "==", "11.0")
            or row_check_version(row, DEVICE_COMPILER, "==", "11.1")
            or row_check_version(row, DEVICE_COMPILER, "==", "11.2")
            or row_check_version(row, DEVICE_COMPILER, "==", "11.3")
        )
        and row_check_name(row, HOST_COMPILER, "==", GCC)
        and row_check_version(row, HOST_COMPILER, "==", "10")
        and row_check_version(row, UBUNTU, "==", "20.04")
    ):
        reason(
            output,
            "gcc-10.3 (provided by Ubuntu 20.04) does not work "
            "together with CUDA 11.0 to 11.3",
        )
        return False

    return True



"""Different support functions.

The row type: Several utile functions to do different checks on the row 
parameters. The type of the row is defined by the pairwise library and the 
parameter types.
Each element of the row have ether the type of Tuple[str, str], where the first 
string is a name and the second string is a version number or the type 
List[Tuple[str, str]], where tuple also stores a name and version combination. 
The list type is only used for the backend values of the `backends` parameter.
The position of the value decides, which parameter a value represent. For 
example the first and second value can be ("gcc", "9"). Only the position 
decides if the first value is the host compiler and second value the device 
compiler or vice versa. 
The global variable `param_map` implements a indirection which allows to 
separate the ordering of the parameters from the algorithm. The `param_map` 
needs to be initialized by the user at each application/test start. 
For Example:
    param_map[HOST_COMPILER] = 0
    param_map[DEVICE_COMPILER] = 1

Example for a row:
    [('gcc', '6'), ('nvcc', '11.4'), 
    [
        ('alpaka_ACC_CPU_B_SEQ_T_SEQ_ENABLE', '1.0.0'), 
        ('alpaka_ACC_CPU_B_OMP2_T_SEQ_ENABLE', '0.0.0'), 
        ('alpaka_ACC_GPU_CUDA_ENABLE', '11.4'), 
        ('alpaka_ACC_GPU_HIP_ENABLE', '0.0.0')
    ], 
    ('cmake', '3.19.8'), ('boost', '1.66.0'), ('alpaka', '0.6.1'), 
    ('ubuntu', '20.04')]
"""

import operator, re, io
from typing import Any, List, Dict, Tuple, Optional, Union
from typeguard import typechecked
from packaging import version as pk_version


from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import

# maps strings to comparision operators
OPERATOR_MAP = {
    "==": operator.eq,
    "!=": operator.ne,
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
}


# no typechecked, because function is performance critical
def strict_equal(first_value: Any, second_value: Any) -> bool:
    """Compare types and values of a and b. If the types are different,
    throw error. If the types are equal the result is a bool.

    Args:
        first_value (Any): first value to compare
        second_value (Any): second value to compare

    Raises:
        TypeError: Is thrown, if the types of a and b are different.

    Returns:
        bool: True if values are equal, otherwise False.
    """
    if not isinstance(first_value, type(second_value)):
        raise TypeError(
            f"a and b has not the same type: {type(first_value)} != {type(second_value)}"
        )

    return first_value == second_value


def is_in_row(row: List, name: str) -> bool:
    """Check if paramater is in the row.

    Args:
        row (List): Row with parameters.
        name (str): The searched parameter.

    Returns:
        bool: Return True, if parameter is in row.
    """
    return param_map[name] < len(row)


# no typechecked, because function is performance critical
def row_check_name(row: List, colum: str, opr: str, name: str) -> bool:
    """Check if colum is in row and if the name matches or not, depending
    of the operator.

    Args:
        row (List): Row to check.
        colum (str): Colum name in the row.
        opr (str): The operator can be "==" (equal) or "!=" (not equal).
        name (str): Name to compare.

    Raises:
        ValueError: Raise error, if operator does not have the value "==" or "!=".

    Returns:
        bool: Return False, if column is not in the row. If the column is in the row,
        return True if the name matches ("==") or not matches ("!=").
    """
    if not opr in ("==", "!="):
        raise ValueError("op (operator) needs to be == or !=")

    return is_in_row(row, colum) and OPERATOR_MAP[opr](
        row[param_map[colum]][NAME], name
    )


# no typechecked, because function is performance critical
def row_check_version(
    row: List,
    colum: str,
    opr: str,
    version: str,
) -> bool:
    """Check if colum is in row and if the version matches or not, depending
    of the operator.

    Args:
        row (List): Row to check.
        colum (str): Colum name in the row.
        opr (str): The operator can be "==", "!=", "<", "<=", ">" and ">=".
        version (str): Version to compare.

    Raises:
        ValueError: Raise error, if operator does not have the supported value.

    Returns:
        bool: Return False, if column is not in the row. If the column is in the row,
        return True if the version.
    """
    if not opr in OPERATOR_MAP:
        raise ValueError(f"operator needs to be: {', '.join(OPERATOR_MAP.keys())}")

    return is_in_row(row, colum) and OPERATOR_MAP[opr](
        pk_version.parse(row[param_map[colum]][VERSION]), pk_version.parse(version)
    )


def backend_is_not_in_row(row: List, backend: str) -> bool:
    """Returns True, if backend is not in backend list.

    Args:
        row (List): Row to check.
        backend (str): Name of the backend, which version should be compared.

    Returns:
        bool: Return True, if backend is not in the row. If backend is in the row and
        the backend name is in the backend list return False else True.
    """

    if not is_in_row(row, BACKENDS):
        return True
    else:
        for row_backend in row[param_map[BACKENDS]]:
            if row_backend[NAME] == backend:
                return False
    return True


# no typechecked, because function is performance critical
def row_check_backend_version(row: List, backend: str, opr: str, version: str) -> bool:
    """Check, if backend exists and if the backend version matches depending of the operator.

    Args:
        row (List): Row to check.
        backend (str): Name of the backend, which version should be compared.
        opr (str): The operator can be "==", "!=", "<", "<=", ">" and ">=".
        version (str): Version to compare.

    Raises:
        ValueError: Raise error, if operator does not have the supported value.

    Returns:
        bool: Return False, if backend does not exist. If the backend name is in the row, return
        True if the version matches.
    """
    if not opr in OPERATOR_MAP:
        raise ValueError(f"operator needs to be: {', '.join(OPERATOR_MAP.keys())}")

    if not is_in_row(row, BACKENDS):
        return False

    for row_backend in row[param_map[BACKENDS]]:
        if row_backend[NAME] == backend:
            return OPERATOR_MAP[opr](
                pk_version.parse(row_backend[VERSION]), pk_version.parse(version)
            )

    return False


@typechecked
def search_and_move_job(
    job_matrix: List[Dict[str, Tuple[str, str]]],
    searched_job: Dict[str, Tuple[str, str]],
    position: int = 0,
) -> bool:
    """Search job, which contains all items of searched_job and move it to list position.

    Args:
        job_matrix (List[Dict[str, Tuple[str, str]]]): The job matrix.
        searched_job (Dict[str, Tuple[str, str]]): Dict of searched items. If all items matches with
        an entry in job list, move entry to position
        position (int, optional): New position of matched entry in job_matrix. Defaults to 0.

    Raises:
        IndexError: Raise error, if searched_job dict is empty.

    Returns:
        bool: True if found and move entry, otherwise False. If False, job_matrix was not modified.
    """
    if len(searched_job) == 0:
        raise IndexError("searched_job must not be empty")

    for index, job_combination in enumerate(job_matrix):
        matched_attributes = 0
        for attribute_name, attribute_value in searched_job.items():
            if (
                attribute_name in job_combination
                and job_combination[attribute_name] == attribute_value
            ):
                matched_attributes += 1
            if matched_attributes == len(searched_job):
                job_matrix.insert(position, job_matrix.pop(index))
                return True
    return False


@typechecked
def filter_job_list(
    job_matrix: List[Dict[str, Dict]], job_name_regex: str
) -> List[Dict[str, Dict]]:
    """Returns list, where all job names matches the job_name_regex.

    Args:
        job_matrix (List[Dict[str, Dict]]): Input job matrix.
        job_name_regex (str): Regex to match. See Python regex.

    Returns:
        List[Dict[str, Dict]]: Filtered job matrix.
    """
    compiled_regex = re.compile(job_name_regex)
    filtered_jobs: List[Dict[str, Dict]] = []

    for job in job_matrix:
        if compiled_regex.match(list(job.keys())[0]):
            filtered_jobs.append(job)

    return filtered_jobs


@typechecked
def reorder_job_list(
    job_matrix: List[Dict[str, Dict]], job_name_regex: str
) -> List[Dict[str, Dict]]:
    """Reorder list with a list of regex. The ordering of the regex in the job_name_regex will be
    also the ordering of the return job matrix.

    For example, the job_name_regex string "^NVCC ^GCC" has the behavior that all NVCC jobs will be
    the first items in the returned job matrix, than all GCC job will follow and than all jobs,
    which does not match the two regex.

    Args:
        job_matrix (List[Dict[str, Dict]]): Input job matrix.
        job_name_regex (str): List of regex, separated by whitespaces. E.g. "^NVCC ^Clang|^GCC ^HIP"

    Returns:
        List[Dict[str, Dict]]: Reordered job matrix.
    """
    # each regex is separated by a whitespace
    ordering_list = job_name_regex.strip().split(" ")

    # reverse list, because reorder_job_list_single_regex() puts matched jobs in the beginning
    ordering_list.reverse()

    tmp_job_matrix = job_matrix
    for regex in ordering_list:
        tmp_job_matrix = reorder_job_list_single_regex(tmp_job_matrix, regex)

    return tmp_job_matrix


@typechecked
def reorder_job_list_single_regex(
    job_matrix: List[Dict[str, Dict]], job_name_regex: str
) -> List[Dict[str, Dict]]:
    """Reorder list with a regex. Put all jobs in the beginning, which names matches the regex. Then
    all other jobs will following.

    Args:
        job_matrix (List[Dict[str, Dict]]): Input job matrix.
        job_name_regex (str): Regex to match. See Python regex.

    Returns:
        List[Dict[str, Dict]]: Reordered job matrix.
    """
    compiled_regex = re.compile(job_name_regex)
    index_list: List[int] = []
    new_job_list: List[Dict[str, Dict]] = []

    for index, job in enumerate(job_matrix):
        if compiled_regex.match(list(job.keys())[0]):
            index_list.append(index)

    for index, job in enumerate(job_matrix):
        if index in index_list:
            new_job_list.insert(0, job)
        else:
            new_job_list.append(job)

    return new_job_list


def reason(output: Optional[Union[io.StringIO, io.TextIOWrapper]], msg: str):
    """Write the message to output if it is not None. This function is used
    in filter functions to print additional information about filter decisions.

    Args:
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): IO object.
        Can be io.StringIO, sys.stdout sys.stderr
        msg (str): the message
    """
    if output:
        print(
            msg,
            file=output,
            end="",
        )



import unittest

from alpaka_job_coverage.filter_compiler_name import general_compiler_filter_typed
from alpaka_job_coverage.filter_compiler_version import compiler_version_filter_typed
from alpaka_job_coverage.filter_backend_version import compiler_backend_filter_typed
from alpaka_job_coverage.filter_software_dependency import (
    software_dependency_filter_typed,
)
from alpaka_job_coverage.globals import *
from utils import manual_version_test


def full_filter_chain(row) -> bool:
    return (
        general_compiler_filter_typed(row)
        and compiler_version_filter_typed(row)
        and compiler_backend_filter_typed(row)
        and software_dependency_filter_typed(row)
    )


class TestHostDeviceCompiler(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[HOST_COMPILER] = 0
        param_map[DEVICE_COMPILER] = 1

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    # test different combinations of NVCC, Clang, GCC and other Compilers as
    # host and device compiler. The version is not important.
    def test_host_device_compiler(self):
        valid_combs = [
            [(GCC, "0"), (NVCC, "0")],
            [(CLANG, "0"), (NVCC, "0")],
            [(CLANG_CUDA, "0"), (CLANG_CUDA, "0")],
        ]

        for comb in valid_combs:
            self.assertTrue(general_compiler_filter_typed(comb))

        invalid_combs = [
            # NVCC is not allowed as host compiler
            [(NVCC, "0"), (NVCC, "0")],
            [(NVCC, "0"), (GCC, "0")],
            [(NVCC, "0"), (CLANG, "0")],
            # only GCC and Clang are allowed as host compiler for nvcc
            [(CLANG_CUDA, "0"), (NVCC, "0")],
            [(HIPCC, "0"), (NVCC, "0")],
        ]

        for comb in invalid_combs:
            self.assertFalse(general_compiler_filter_typed(comb))


class TestGeneralFilterFunctionality(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[HOST_COMPILER] = 0
        param_map[DEVICE_COMPILER] = 1
        param_map[BACKENDS] = 2
        # UBUNTU and CXX_STANDARD are only required, because
        # software_dependency_filter_typed do a look up in param_map for it
        param_map[CXX_STANDARD] = 3
        param_map[UBUNTU] = 4

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    def test_if_backend_is_not_set(self):
        # general_compiler_filter_typed should be always return true, because
        # GCC is a valid host compiler and NVCC a valid device compiler

        # compiler_backend_filter_typed should always return true for the valid
        # combination, because if the backend parameter
        # (ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2") is added, it becomes a valid
        # combination

        # compiler_version_filter_typed filters invalid combination depending of
        # host and device compiler version

        # no rule in software_dependency_filter_typed should affect this test

        comb_valid_1 = [(GCC, "9"), (NVCC, "11.2")]

        self.assertTrue(general_compiler_filter_typed(comb_valid_1))
        self.assertTrue(compiler_backend_filter_typed(comb_valid_1))
        self.assertTrue(compiler_version_filter_typed(comb_valid_1))
        self.assertTrue(software_dependency_filter_typed(comb_valid_1))
        self.assertTrue(full_filter_chain(comb_valid_1))

        comb_invalid_1 = [(GCC, "13"), (NVCC, "11.2")]

        self.assertTrue(general_compiler_filter_typed(comb_invalid_1))
        self.assertTrue(compiler_backend_filter_typed(comb_invalid_1))
        self.assertFalse(compiler_version_filter_typed(comb_invalid_1))
        self.assertTrue(software_dependency_filter_typed(comb_invalid_1))
        self.assertFalse(full_filter_chain(comb_invalid_1))

        comb_valid_2 = [(CLANG, "9"), (NVCC, "11.2")]

        self.assertTrue(general_compiler_filter_typed(comb_valid_2))
        self.assertTrue(compiler_backend_filter_typed(comb_valid_2))
        self.assertTrue(compiler_version_filter_typed(comb_valid_2))
        self.assertTrue(software_dependency_filter_typed(comb_valid_2))
        self.assertTrue(full_filter_chain(comb_valid_2))

        comb_invalid_2 = [(CLANG, "17"), (NVCC, "11.2")]

        self.assertTrue(general_compiler_filter_typed(comb_invalid_2))
        self.assertTrue(compiler_backend_filter_typed(comb_invalid_2))
        self.assertFalse(compiler_version_filter_typed(comb_invalid_2))
        self.assertTrue(software_dependency_filter_typed(comb_invalid_2))
        self.assertFalse(full_filter_chain(comb_invalid_2))

    def test_if_backend_is_disabled(self):
        # because the backend is off, compiler_backend_filter_typed should return
        # False every time

        comb_valid_1 = [
            (GCC, "9"),
            (NVCC, "11.2"),
            [(ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER)],
        ]

        self.assertTrue(general_compiler_filter_typed(comb_valid_1))
        self.assertFalse(compiler_backend_filter_typed(comb_valid_1))
        self.assertTrue(compiler_version_filter_typed(comb_valid_1))
        self.assertTrue(software_dependency_filter_typed(comb_valid_1))
        self.assertFalse(full_filter_chain(comb_valid_1))

        comb_valid_2 = [
            (GCC, "9"),
            (NVCC, "11.2"),
            [
                (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                (ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER),
            ],
        ]

        self.assertTrue(general_compiler_filter_typed(comb_valid_2))
        self.assertFalse(compiler_backend_filter_typed(comb_valid_2))
        self.assertTrue(compiler_version_filter_typed(comb_valid_2))
        self.assertTrue(software_dependency_filter_typed(comb_valid_2))
        self.assertFalse(full_filter_chain(comb_valid_2))

        comb_valid_3 = [
            (CLANG, "9"),
            (NVCC, "11.2"),
            [
                (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                (ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER),
            ],
        ]

        self.assertTrue(general_compiler_filter_typed(comb_valid_3))
        self.assertFalse(compiler_backend_filter_typed(comb_valid_3))
        self.assertTrue(compiler_version_filter_typed(comb_valid_3))
        self.assertTrue(software_dependency_filter_typed(comb_valid_3))
        self.assertFalse(full_filter_chain(comb_valid_3))

        comb_invalid_1 = [
            (GCC, "13"),
            (NVCC, "11.2"),
            [(ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER)],
        ]

        self.assertTrue(general_compiler_filter_typed(comb_invalid_1))
        self.assertFalse(compiler_backend_filter_typed(comb_invalid_1))
        self.assertFalse(compiler_version_filter_typed(comb_invalid_1))
        self.assertTrue(software_dependency_filter_typed(comb_invalid_1))
        self.assertFalse(full_filter_chain(comb_invalid_1))

        comb_invalid_2 = [
            (GCC, "13"),
            (NVCC, "11.2"),
            [
                (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                (ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER),
            ],
        ]

        self.assertTrue(general_compiler_filter_typed(comb_invalid_2))
        self.assertFalse(compiler_backend_filter_typed(comb_invalid_2))
        self.assertFalse(compiler_version_filter_typed(comb_invalid_2))
        self.assertTrue(software_dependency_filter_typed(comb_invalid_2))
        self.assertFalse(full_filter_chain(comb_invalid_2))

        comb_invalid_3 = [
            (CLANG, "13"),
            (NVCC, "11.2"),
            [
                (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                (ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER),
            ],
        ]

        self.assertTrue(general_compiler_filter_typed(comb_invalid_3))
        self.assertFalse(compiler_backend_filter_typed(comb_invalid_3))
        self.assertFalse(compiler_version_filter_typed(comb_invalid_3))
        self.assertTrue(software_dependency_filter_typed(comb_invalid_3))
        self.assertFalse(full_filter_chain(comb_invalid_3))

    def test_backend_versions(self):
        # for NVCC, the compiler version needs to be equal to the backend version
        for version in ["10.2", "11.0", "12.2"]:
            comb1 = [
                (GCC, "9"),
                (NVCC, version),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, version)],
            ]
            self.assertTrue(compiler_backend_filter_typed(comb1))

            comb2 = [
                (CLANG, "9"),
                (NVCC, version),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, version)],
            ]
            self.assertTrue(compiler_backend_filter_typed(comb2))

            comb3 = [
                (GCC, "9"),
                (NVCC, version),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_GPU_CUDA_ENABLE, version),
                ],
            ]
            self.assertTrue(compiler_backend_filter_typed(comb3))

            comb4 = [
                (CLANG, "9"),
                (NVCC, version),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_GPU_CUDA_ENABLE, version),
                ],
            ]
            self.assertTrue(compiler_backend_filter_typed(comb4))

        for nvcc_version, backend_version in [
            ("10.1", "10.2"),
            ("11.0", "11.5"),
            ("11.8", "12.2"),
        ]:
            comb1 = [
                (GCC, "9"),
                (NVCC, nvcc_version),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, backend_version)],
            ]
            self.assertFalse(compiler_backend_filter_typed(comb1))

            comb2 = [
                (CLANG, "9"),
                (NVCC, nvcc_version),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, backend_version)],
            ]
            self.assertFalse(compiler_backend_filter_typed(comb2))

            comb3 = [
                (GCC, "9"),
                (NVCC, nvcc_version),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_GPU_CUDA_ENABLE, backend_version),
                ],
            ]
            self.assertFalse(compiler_backend_filter_typed(comb3))

            comb4 = [
                (CLANG, "9"),
                (NVCC, nvcc_version),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_GPU_CUDA_ENABLE, backend_version),
                ],
            ]
            self.assertFalse(compiler_backend_filter_typed(comb4))


class TestNvccGccCompatibility(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[DEVICE_COMPILER] = 0
        param_map[HOST_COMPILER] = 1

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    # General test of the algorithm, which should decide if a GCC and CUDA SDK
    # version combination is valid.
    def test_general_algorithm(self):
        expected_results = [
            ("10.0", "6", True),
            ("10.0", "7", True),
            ("10.0", "8", False),
            ("10.0", "13", False),
            ("10.1", "6", True),
            ("10.1", "7", True),
            ("10.1", "8", True),
            ("10.1", "13", False),
            ("10.2", "6", True),
            ("10.2", "7", True),
            ("10.2", "8", True),
            ("10.2", "13", False),
            ("11.0", "6", True),
            ("11.0", "9", True),
            ("11.0", "10", False),
            ("11.0", "12", False),
        ]

        for nvcc_version, gcc_version, expected_value in expected_results:
            self.assertEqual(
                compiler_version_filter_typed(
                    [(NVCC, nvcc_version), (GCC, gcc_version)]
                ),
                expected_value,
                f"nvcc {nvcc_version} + gcc {gcc_version} -> {expected_value}",
            )

    # Test the combination of GCC version and CUDA SDK version.
    # Validate the combinations, which are defined in the release notes of the
    # CUDA SDK.
    def test_known_version(self):
        expected_results = [
            ("10.0", "7", True),
            ("10.0", "8", False),
            ("10.1", "8", True),
            ("10.1", "9", False),
            ("10.2", "8", True),
            ("10.2", "9", False),
            ("11.0", "9", True),
            ("11.0", "10", False),
            ("11.1", "10", True),
            ("11.1", "11", False),
            ("11.2", "10", True),
            ("11.2", "11", False),
            ("11.3", "10", True),
            ("11.3", "11", False),
            ("11.4", "11", True),
            ("11.4", "12", False),
            ("11.5", "11", True),
            ("11.5", "12", False),
            ("11.6", "11", True),
            ("11.6", "12", False),
            ("11.7", "11", True),
            ("11.7", "12", False),
            ("11.8", "11", True),
            ("11.8", "12", False),
            ("12.0", "12", True),
            ("12.0", "13", False),
            ("12.1", "12", True),
            ("12.1", "13", False),
            ("12.2", "12", True),
            ("12.2", "13", False),
        ]

        for nvcc_version, gcc_version, expected_value in expected_results:
            self.assertEqual(
                compiler_version_filter_typed(
                    [(NVCC, nvcc_version), (GCC, gcc_version)]
                ),
                expected_value,
                f"nvcc {nvcc_version} + gcc {gcc_version} -> {expected_value}",
            )

    # Test if a unknown CUDA SDK version returns always true.
    # This avoid behavior avoids, that valid combinations are silently disabled
    # if a new CUDA SDK version is released.
    def test_unknown_version(self):
        expected_results = [
            ("42.0", "45", True),
        ]

        # verify that our test version are not supported
        manual_version_test(self, NVCC, [], ["42.0"])
        manual_version_test(self, GCC, [], ["45"])

        for nvcc_version, gcc_version, expected_value in expected_results:
            self.assertEqual(
                compiler_version_filter_typed(
                    [(NVCC, nvcc_version), (GCC, gcc_version)]
                ),
                expected_value,
                f"nvcc {nvcc_version} + gcc {gcc_version} -> {expected_value}",
            )


class TestNvccClangCompatibility(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[DEVICE_COMPILER] = 0
        param_map[HOST_COMPILER] = 1

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    # Test the combination of Clang version and CUDA SDK version.
    # Validate the combinations, which are defined in the release notes of the
    # CUDA SDK.
    def test_known_version(self):
        expected_results = [
            ("10.0", "6", True),
            ("10.0", "7", False),
            ("10.1", "8", True),
            ("10.1", "9", False),
            ("10.2", "8", True),
            ("10.2", "9", False),
            ("11.0", "9", True),
            ("11.0", "10", False),
            ("11.1", "10", True),
            ("11.1", "11", False),
            ("11.2", "11", True),
            ("11.2", "12", False),
            # because of compiler bugs, clang is disabled for CUDA 11.3 until 11.5
            ("11.3", "11", False),
            ("11.3", "12", False),
            ("11.4", "12", False),
            ("11.4", "13", False),
            ("11.5", "12", False),
            ("11.5", "13", False),
            ("11.6", "13", True),
            ("11.6", "14", False),
            ("11.7", "13", True),
            ("11.7", "14", False),
            ("11.8", "13", True),
            ("11.8", "14", False),
            ("12.0", "14", True),
            ("12.0", "15", False),
            ("12.1", "15", True),
            ("12.1", "16", False),
            ("12.2", "15", True),
            ("12.2", "16", False),
        ]

        for nvcc_version, clang_version, expected_value in expected_results:
            self.assertEqual(
                compiler_version_filter_typed(
                    [(NVCC, nvcc_version), (CLANG, clang_version)]
                ),
                expected_value,
                f"nvcc {nvcc_version} + clang {clang_version} -> {expected_value}",
            )

    # Test if a unknown CUDA SDK version returns always true.
    # This avoid behavior avoids, that valid combinations are silently disabled
    # if a new CUDA SDK version is released.
    def test_unknown_version(self):
        expected_results = [
            ("42.0", "45", True),
        ]

        # verify that our test version are not supported
        manual_version_test(self, NVCC, [], ["42.0"])
        manual_version_test(self, CLANG, [], ["45"])

        for nvcc_version, clang_version, expected_value in expected_results:
            self.assertEqual(
                compiler_version_filter_typed(
                    [(NVCC, nvcc_version), (CLANG, clang_version)]
                ),
                expected_value,
                f"nvcc {nvcc_version} + clang {clang_version} -> {expected_value}",
            )


class TestNvccCxxStandard(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[DEVICE_COMPILER] = 0
        param_map[CXX_STANDARD] = 1
        # the following parameters are not used in this test
        param_map[HOST_COMPILER] = 2
        param_map[BACKENDS] = 3
        param_map[UBUNTU] = 4

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    # Test the combination of NVCC as device compiler an the C++ standard.
    # If SDK version is unknown, all C++ standards should be allowed.
    def test_nvcc_cxx(self):
        # verify that our test version are not supported
        manual_version_test(self, NVCC, [], ["12.2, 45.0"])

        for nvcc_version, max_cxx in [
            ("11.0", 17),
            ("11.2", 17),
            ("11.8", 17),
            ("12.0", 20),
            ("12.2", 20),
            # not released version
            # therefore they should support all C++ versions
            ("12.3", 32),
            ("45.0", 32),
        ]:
            for cxx_version in [11, 14, 17, 20, 23, 26, 29, 32]:
                comb = [(NVCC, nvcc_version), (CXX_STANDARD, str(cxx_version))]
                if cxx_version <= max_cxx:
                    self.assertTrue(
                        software_dependency_filter_typed(comb),
                        f"NVCC {nvcc_version} + CXX {cxx_version}",
                    )
                else:
                    self.assertFalse(
                        software_dependency_filter_typed(comb),
                        f"NVCC {nvcc_version} + CXX {cxx_version}",
                    )


# tests for Clang as CUDA compiler
class TestClangCUDA(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[HOST_COMPILER] = 0
        param_map[DEVICE_COMPILER] = 1
        param_map[BACKENDS] = 2
        # the following parameters are not used in this test
        param_map[CXX_STANDARD] = 3
        param_map[UBUNTU] = 4
        param_map[CMAKE] = 5

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    # standalone CLANG_CUDA as host compiler should pass all tests
    def test_host_compiler_name(self):
        comb = [(CLANG_CUDA, "99")]

        self.assertTrue(general_compiler_filter_typed(comb))
        self.assertTrue(compiler_version_filter_typed(comb))
        self.assertTrue(compiler_backend_filter_typed(comb))
        self.assertTrue(software_dependency_filter_typed(comb))
        self.assertTrue(full_filter_chain(comb))

    # CLANG_CUDA as host compiler can be only used with CLANG_CUDA as device
    # compiler an vice vera
    def test_host_device_compiler_name(self):
        valid_comb = [(CLANG_CUDA, "99"), (CLANG_CUDA, "99")]

        self.assertTrue(general_compiler_filter_typed(valid_comb))
        self.assertTrue(compiler_version_filter_typed(valid_comb))
        self.assertTrue(compiler_backend_filter_typed(valid_comb))
        self.assertTrue(software_dependency_filter_typed(valid_comb))
        self.assertTrue(full_filter_chain(valid_comb))

        invalid_combs = [
            [(CLANG_CUDA, "99"), (GCC, "99")],
            [(CLANG_CUDA, "99"), (CLANG, "99")],
            [(CLANG_CUDA, "99"), (HIPCC, "99")],
            [(CLANG_CUDA, "99"), (NVCC, "99")],
            [(GCC, "99"), (CLANG_CUDA, "99")],
            [(CLANG, "99"), (CLANG_CUDA, "99")],
            [(HIPCC, "99"), (CLANG_CUDA, "99")],
            [(NVCC, "99"), (CLANG_CUDA, "99")],
        ]

        for comb in invalid_combs:
            self.assertFalse(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )
            self.assertFalse(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]}, DEVICE_COMPILER: {comb[1][0]}",
            )

    # host and device compiler needs to have the same version number
    def test_host_device_compiler_version(self):
        valid_combs = [
            [(CLANG_CUDA, "14"), (CLANG_CUDA, "14")],
            [(CLANG_CUDA, "16"), (CLANG_CUDA, "16")],
            [(CLANG_CUDA, "20"), (CLANG_CUDA, "20")],
            [(CLANG_CUDA, "99"), (CLANG_CUDA, "99")],
        ]

        for comb in valid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )

        invalid_combs = [
            [(CLANG_CUDA, "16"), (CLANG_CUDA, "17")],
            [(CLANG_CUDA, "21"), (CLANG_CUDA, "20")],
            # only Clang 14 and newer are supported as CUDA compiler
            [(CLANG_CUDA, "9"), (CLANG_CUDA, "9")],
            [(CLANG_CUDA, "13"), (CLANG_CUDA, "13")],
        ]

        for comb in invalid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertFalse(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )
            self.assertFalse(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}",
            )

    # the CUDA backend needs to enabled and Clang needs to support the CUDA
    # SDK
    def test_host_device_compiler_backend(self):
        valid_combs = [
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "11.5")],
            ],
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2")],
            ],
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE, OFF_VER),
                    (ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2"),
                    (ALPAKA_ACC_GPU_HIP_ENABLE, OFF_VER),
                ],
            ],
        ]

        for comb in valid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )

        invalid_combs = [
            # CUDA backend needs to be enabled
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER)],
            ],
            # CUDA backend needs to be supported
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "12.0")],
            ],
            # HIP backend cannot be enabled, when CLANG_CUDA is used
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [
                    (ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                    (ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE, OFF_VER),
                    (ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2"),
                    (ALPAKA_ACC_GPU_HIP_ENABLE, "5.0"),
                ],
            ],
        ]

        for comb in invalid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertFalse(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )
            self.assertFalse(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"BACKENDS: {comb[2]}",
            )

    # test all supported Clang CUDA SDK combinations
    def test_all_clang_cuda_version_combinations(self):
        cuda_versions = [
            9.0,
            9.1,
            9.2,
            10.0,
            10.1,
            10.2,
            11.0,
            11.1,
            11.2,
            11.3,
            11.4,
            11.5,
            11.6,
            11.7,
            11.8,
            12.0,
            12.1,
        ]

        # key: clang version
        # value: max supported CUDA SDK
        supported_versions = {
            7: 9.2,
            8: 10,
            9: 10.1,
            10: 10.1,
            11: 11.0,
            12: 11.0,
            13: 11.2,
            14: 11.5,
            15: 11.5,
            16: 11.5,
        }

        for clang_version in supported_versions.keys():
            for cuda_version in cuda_versions:
                comb = [
                    (CLANG_CUDA, str(clang_version)),
                    (CLANG_CUDA, str(clang_version)),
                    [(ALPAKA_ACC_GPU_CUDA_ENABLE, str(cuda_version))],
                ]

                if cuda_version > supported_versions[clang_version]:
                    self.assertFalse(
                        compiler_backend_filter_typed(comb),
                        f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                        f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                        f"BACKENDS: {comb[2]}",
                    )
                else:
                    self.assertTrue(
                        compiler_backend_filter_typed(comb),
                        f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                        f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                        f"BACKENDS: {comb[2]}",
                    )

        # test if combination of non released clang version and released CUDA
        # SDK is allowed
        self.assertTrue(
            compiler_backend_filter_typed(
                [
                    (CLANG_CUDA, "30"),
                    (CLANG_CUDA, "30"),
                    [(ALPAKA_ACC_GPU_CUDA_ENABLE, "11.5")],
                ]
            ),
            f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
            f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
            f"BACKENDS: {comb[2]}",
        )

        # test if combination of non released clang version and non released
        # CUDA SDK is allowed
        self.assertTrue(
            compiler_backend_filter_typed(
                [
                    (CLANG_CUDA, "30"),
                    (CLANG_CUDA, "30"),
                    [(ALPAKA_ACC_GPU_CUDA_ENABLE, "30.0")],
                ]
            ),
            f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
            f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
            f"BACKENDS: {comb[2]}",
        )

        # test if combination of released clang version and non released
        # CUDA SDK is not allowed
        self.assertFalse(
            compiler_backend_filter_typed(
                [
                    (CLANG_CUDA, "15"),
                    (CLANG_CUDA, "15"),
                    [(ALPAKA_ACC_GPU_CUDA_ENABLE, "30.0")],
                ]
            ),
            f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
            f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
            f"BACKENDS: {comb[2]}",
        )

    # Clang 9 and older does not support C++ 20.
    def test_cxx_support(self):
        for clang_version in range(7, 16):
            for cxx_version in [17, 20]:
                comb = [
                    (CLANG_CUDA, str(clang_version)),
                    (CLANG_CUDA, str(clang_version)),
                    [(ALPAKA_ACC_GPU_CUDA_ENABLE, "9.2")],
                    (CXX_STANDARD, str(cxx_version)),
                ]
                self.assertTrue(
                    general_compiler_filter_typed(comb),
                    f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                    f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                    f"C++: {comb[3]}",
                )

                # does not test compiler_version_filter_typed(), because of the filter
                # rule, that Clang 13 and older is not allowed as CUDA compiler

                self.assertTrue(
                    compiler_backend_filter_typed(comb),
                    f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                    f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                    f"C++: {comb[3]}",
                )

                if clang_version <= 9 and cxx_version > 17:
                    self.assertFalse(
                        software_dependency_filter_typed(comb),
                        f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                        f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                        f"C++: {comb[3]}",
                    )
                    self.assertFalse(
                        full_filter_chain(comb),
                        f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                        f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                        f"C++: {comb[3]}",
                    )
                else:
                    self.assertTrue(
                        software_dependency_filter_typed(comb),
                        f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                        f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                        f"C++: {comb[3]}",
                    )

    # Clang 11 and 12 is not available in the Ubuntu 18.04 ppa
    def test_ubuntu1804_ppa(self):
        for clang_version in range(7, 16):
            comb = [
                (CLANG_CUDA, str(clang_version)),
                (CLANG_CUDA, str(clang_version)),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "9.2")],
                (CXX_STANDARD, "17"),
                (UBUNTU, "18.04"),
            ]

            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"Ubuntu: {comb[4]}",
            )

            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"Ubuntu: {comb[4]}",
            )

            if clang_version == 11 or clang_version == 12:
                self.assertFalse(
                    software_dependency_filter_typed(comb),
                    f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                    f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                    f"Ubuntu: {comb[4]}",
                )
                self.assertFalse(
                    full_filter_chain(comb),
                    f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                    f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                    f"Ubuntu: {comb[4]}",
                )
            else:
                self.assertTrue(
                    software_dependency_filter_typed(comb),
                    f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                    f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                    f"Ubuntu: {comb[4]}",
                )

    # CMake 3.18 and older does not support Clang as CUDA compiler
    def test_cmake_version(self):
        valid_combs = [
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2")],
                (CXX_STANDARD, "17"),
                (UBUNTU, "20.04"),
                (CMAKE, "3.19"),
            ],
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2")],
                (CXX_STANDARD, "17"),
                (UBUNTU, "20.04"),
                (CMAKE, "3.22"),
            ],
        ]

        for comb in valid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )
            self.assertTrue(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )
            self.assertTrue(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )

        invalid_combs = [
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2")],
                (CXX_STANDARD, "17"),
                (UBUNTU, "20.04"),
                (CMAKE, "3.18"),
            ],
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2")],
                (CXX_STANDARD, "17"),
                (UBUNTU, "20.04"),
                (CMAKE, "3.17"),
            ],
            [
                (CLANG_CUDA, "14"),
                (CLANG_CUDA, "14"),
                [(ALPAKA_ACC_GPU_CUDA_ENABLE, "11.2")],
                (CXX_STANDARD, "17"),
                (UBUNTU, "20.04"),
                (CMAKE, "3.18.3"),
            ],
        ]

        for comb in invalid_combs:
            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )
            self.assertTrue(
                compiler_version_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )
            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )
            self.assertFalse(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )
            self.assertFalse(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"CMAKE: {comb[5][1]}",
            )



import unittest

from alpaka_job_coverage.util import backend_is_not_in_row
from alpaka_job_coverage.globals import *

# the shape of the row function parameter is described in the module
# documentation of alpaka_job_coverage.util


class TestBackendIsNotInRowBackendsOnly(unittest.TestCase):
    def setUp(self):
        global param_map
        param_map[BACKENDS] = 0

    def setDown(self):
        global param_map
        param_map = {}

    def test_empty_backend_list(self):
        self.assertTrue(backend_is_not_in_row([], ALPAKA_ACC_GPU_CUDA_ENABLE))

    def test_is_not_in_filled_list_single_element(self):
        self.assertTrue(
            backend_is_not_in_row(
                [[(ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER)]],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )

    def test_is_not_in_filled_list_multi_element(self):
        self.assertTrue(
            backend_is_not_in_row(
                [
                    [
                        (ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER),
                        (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                        (ALPAKA_ACC_GPU_HIP_ENABLE, "5.5"),
                    ],
                ],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )

    def test_is_in_filled_list_multi_element(self):
        self.assertFalse(
            backend_is_not_in_row(
                [
                    [
                        (ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER),
                        (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                        (ALPAKA_ACC_GPU_HIP_ENABLE, "5.5"),
                        (ALPAKA_ACC_GPU_CUDA_ENABLE, "11.4"),
                    ],
                ],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )

        self.assertFalse(
            backend_is_not_in_row(
                [
                    [
                        (ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER),
                        (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                        (ALPAKA_ACC_GPU_CUDA_ENABLE, OFF_VER),
                        (ALPAKA_ACC_GPU_HIP_ENABLE, "5.5"),
                    ],
                ],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )


class TestBackendIsNotInRowMixed(unittest.TestCase):
    def setUp(self):
        global param_map
        param_map[HOST_COMPILER] = 0
        param_map[DEVICE_COMPILER] = 1
        param_map[BACKENDS] = 2
        param_map[CMAKE] = 3

    def setDown(self):
        global param_map
        param_map = {}

    def test_empty_backend_list(self):
        self.assertTrue(backend_is_not_in_row([], ALPAKA_ACC_GPU_CUDA_ENABLE))

    def test_backend_not_included(self):
        self.assertTrue(
            backend_is_not_in_row([(GCC, "9"), (GCC, "9")], ALPAKA_ACC_GPU_CUDA_ENABLE)
        )

    def test_single_backend(self):
        self.assertTrue(
            backend_is_not_in_row(
                [(GCC, "9"), (GCC, "9"), [(ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER)]],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )

        self.assertFalse(
            backend_is_not_in_row(
                [(GCC, "9"), (GCC, "9"), [(ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER)]],
                ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE,
            )
        )

    def test_multiple_backend(self):
        self.assertTrue(
            backend_is_not_in_row(
                [
                    (GCC, "9"),
                    (GCC, "9"),
                    [
                        (ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER),
                        (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                        (ALPAKA_ACC_GPU_HIP_ENABLE, "5.5"),
                    ],
                ],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )

        self.assertFalse(
            backend_is_not_in_row(
                [
                    (GCC, "9"),
                    (GCC, "9"),
                    [
                        (ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER),
                        (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                        (ALPAKA_ACC_GPU_HIP_ENABLE, "5.5"),
                        (ALPAKA_ACC_GPU_CUDA_ENABLE, "11.4"),
                    ],
                ],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )

    def test_multiple_backends_with_following_parameter(self):
        self.assertTrue(
            backend_is_not_in_row(
                [
                    (GCC, "9"),
                    (GCC, "9"),
                    [
                        (ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER),
                        (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                        (ALPAKA_ACC_GPU_HIP_ENABLE, "5.5"),
                    ],
                    (CMAKE, "3.25"),
                ],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )

        self.assertFalse(
            backend_is_not_in_row(
                [
                    (GCC, "9"),
                    (GCC, "9"),
                    [
                        (ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE, ON_VER),
                        (ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE, ON_VER),
                        (ALPAKA_ACC_GPU_HIP_ENABLE, "5.5"),
                        (ALPAKA_ACC_GPU_CUDA_ENABLE, "11.4"),
                    ],
                    (CMAKE, "3.25"),
                ],
                ALPAKA_ACC_GPU_CUDA_ENABLE,
            )
        )



"""Filter rules basing on host and device compiler names.
"""

import io

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_job_coverage.util import row_check_name, is_in_row, reason
from typing import List, Tuple, Union, Optional
from typeguard import typechecked


def get_required_parameter() -> List[str]:
    """Returns a list of parameters which are required for using the filter defined by this module.
    Returns:
        List[str]: list of parameters
    """
    return [HOST_COMPILER, DEVICE_COMPILER]


@typechecked
def general_compiler_filter_typed(
    row: List[Union[Tuple[str, str], List[Tuple[str, str]]]],
    output: Optional[Union[io.StringIO, io.TextIOWrapper]] = None,
) -> bool:
    """Type checked version of general_compiler_filter(). Should be only used for
    testing or tooling. The type check adds a big overhead, which slows down
    pair-wise generator by the factor 30.

    Args:
        row (List[Union[Tuple[str, str], List[Tuple[str, str]]]]): Combination
        to verify. The row can contain up to all combination fields and at least
        two items.
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): Write
        additional information about filter decisions to the IO object
        (io.SringIO, sys.stdout, sys.stderr). If it is None no information is
        generated.

    Returns:
        bool: True, if combination is valid, otherwise False.
    """
    return general_compiler_filter(row, output)


def general_compiler_filter(
    row: List, output: Optional[Union[io.StringIO, io.TextIOWrapper]] = None
) -> bool:
    """Filter rules basing on host and device compiler names.

    Args:
        row (List): Combination to verify. The row can contain
        up to all combination fields and at least two items.
        output (Optional[Union[io.StringIO, io.TextIOWrapper]]): Write
        additional information about filter decisions to the IO object
        (io.SringIO, sys.stdout, sys.stderr). If it is None no information is
        generated.

    Returns:
        bool: True, if combination is valid, otherwise False.
    """

    # it is not allow to use the nvcc as host compiler
    if row_check_name(row, HOST_COMPILER, "==", NVCC):
        reason(output, "nvcc is not allowed as host compiler")
        return False

    # only the nvcc allows to combine different host and device compiler
    if (
        is_in_row(row, HOST_COMPILER)
        and is_in_row(row, DEVICE_COMPILER)
        and (
            row[param_map[DEVICE_COMPILER]][NAME] != NVCC
            and row[param_map[HOST_COMPILER]][NAME]
            != row[param_map[DEVICE_COMPILER]][NAME]
        )
    ):
        reason(output, "host and device compiler must be the same (except for nvcc)")
        return False

    # only clang and gcc are allowed as nvcc host compiler
    if row_check_name(row, DEVICE_COMPILER, "==", NVCC) and not (
        row_check_name(row, HOST_COMPILER, "==", GCC)
        or row_check_name(row, HOST_COMPILER, "==", CLANG)
    ):
        reason(output, "only clang and gcc are allowed as nvcc host compilers")
        return False

    return True



import unittest

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_job_coverage.filter_compiler_name import general_compiler_filter_typed


# Test all compiler names except nvcc and clang-cuda. This is tested in
# test_cuda_sdk.py
class TestHostDeviceCompiler(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[HOST_COMPILER] = 0
        param_map[DEVICE_COMPILER] = 1

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    def test_same_host_device_compiler(self):
        valid_combs = [
            [(GCC, "0"), (GCC, "0")],
            [(CLANG, "0"), (CLANG, "0")],
            [(HIPCC, "0"), (HIPCC, "0")],
            [(ICPX, "0"), (ICPX, "0")],
        ]

        for comb in valid_combs:
            self.assertTrue(general_compiler_filter_typed(comb))

        invalid_combs = [
            [(GCC, "0"), (CLANG, "0")],
            [(GCC, "0"), (CLANG, "0")],
            [(GCC, "0"), (HIPCC, "0")],
            [(HIPCC, "0"), (CLANG, "0")],
            [(GCC, "0"), (ICPX, "0")],
            [(CLANG, "0"), (ICPX, "0")],
            [(ICPX, "0"), (CLANG, "0")]
        ]

        for comb in invalid_combs:
            self.assertFalse(general_compiler_filter_typed(comb))



import setuptools
import os


# the setup.py is executed in an environment, therefore files can not be open via relative
cwd = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# attention, the version.txt needs to be added to the MANIFEST.in, otherwise the file is not
# available during build
with open("version.txt", "r", encoding="utf-8") as fh:
    version = fh.read()

setuptools.setup(
    name="alpaka-job-coverage",
    version=version,
    author="Simeon Ehrig",
    author_email="s.ehrig@hzdr.de",
    description="The library provides everything needed to generate a sparse combination matrix for alpaca-based projects, including a set of general-purpose combination rules.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alpaka-group/alpaka-job-matrix-library",
    project_urls={
        "Bug Tracker": "https://github.com/alpaka-group/alpaka-job-matrix-library/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    entry_points={
        "console_scripts": ["ajc-validate=alpaka_job_coverage.validate:main"]
    },
    python_requires=">=3.8",
    install_requires=[
        "allpairspy == 2.5.0",
        "typeguard",
        "pyaml",
        "types-PyYAML",
        "packaging",
    ],
)



import unittest

from alpaka_job_coverage.filter_compiler_name import general_compiler_filter_typed
from alpaka_job_coverage.filter_compiler_version import compiler_version_filter_typed
from alpaka_job_coverage.filter_backend_version import compiler_backend_filter_typed
from alpaka_job_coverage.filter_software_dependency import (
    software_dependency_filter_typed,
)
from alpaka_job_coverage.globals import *

# The files contains tests for single, specific rules


def full_filter_chain(row) -> bool:
    return (
        general_compiler_filter_typed(row)
        and compiler_version_filter_typed(row)
        and compiler_backend_filter_typed(row)
        and software_dependency_filter_typed(row)
    )


class TestNVCC11GCC103UBUNTU2004(unittest.TestCase):
    def setUp(self):
        global param_map
        # set param_map, that filters expect the following parameters in the
        # order
        param_map[HOST_COMPILER] = 0
        param_map[DEVICE_COMPILER] = 1
        param_map[UBUNTU] = 2
        param_map[CXX_STANDARD] = 3

    def setDown(self):
        global param_map
        # reset param_map for following up tests
        param_map = {}

    # test if nvcc 11.3 + gcc 10 + Ubuntu 20.04 is disabled
    def test_forbid_combination(self):
        for nvcc_version in ["11.0", "11.1", "11.2", "11.3"]:
            comb = [(GCC, "10"), (NVCC, nvcc_version), (UBUNTU, "20.04")]

            self.assertTrue(
                general_compiler_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"UBUNTU: {comb[2][1]}",
            )

            # nvcc 11.0 does not support GCC 10
            if nvcc_version != "11.0":
                self.assertTrue(
                    compiler_version_filter_typed(comb),
                    f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                    f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                    f"UBUNTU: {comb[2][1]}",
                )
            else:
                self.assertFalse(
                    compiler_version_filter_typed(comb),
                    f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                    f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                    f"UBUNTU: {comb[2][1]}",
                )

            self.assertTrue(
                compiler_backend_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"UBUNTU: {comb[2][1]}",
            )
            self.assertFalse(
                software_dependency_filter_typed(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"UBUNTU: {comb[2][1]}",
            )
            self.assertFalse(
                full_filter_chain(comb),
                f"HOST_COMPILER: {comb[0][0]} {comb[0][1]}, "
                f"DEVICE_COMPILER: {comb[1][0]} {comb[1][1]}, "
                f"UBUNTU: {comb[2][1]}",
            )
